﻿import math
import numpy as np
"""
Набор классов для вычисления вероятностей n-грамм слов. 
Базовым является класс NGramsCounter, который подсчитывает число
n-грамм от 1 (одиночные слова) до order (последовательность order слов).
Входом является строка (тогда n-граммы будут по символам) или список слов.

counter = NGramsCounter(3)                      # будет 1,2,3 - грамм
counter.add('мама мыла раму')                   # подсчитываем n-граммы букв

print("размер словаря:", len(counter.root))     # 7
print("слова  словаря:", counter.branches())    # [('м', 4), ('а', 4), ...]
print(counter.ngrams[1:])                       # [14, 13, 12] число n-грамм n=1,2,3
print(counter.unique() )                        # [7, 10, 12] уникальных n-грамм

N1, N2, _ = counter.counts('м')                 # N1 раз была 'м', N2 - длина текста 
print("N('м'), N       = ", N1, N2)             # 4, 14 
print("P('м')          = ", counter.prob('м'))  # 0.2857 = 4/14 = N1/N2

N1, N2, _ = counter.counts('ма')                # N1 раз было 'ма', N2 - было 'м'
print("N('ма'),N('м')  = ", N1, N2)             # 2, 4 
print("P('м=>а')       = ", counter.prob('ма')) # 0.5 = 2/4 = N1/N2

print("P(ма=>м) = ", counter.prob('мам'))       # 0.5   = 2/4  условная 
print("P(мам)   = ", counter.prob('мам', False))# 0.083 = 1/12 совместная

print(counter.branches('ма'))                   # [('м', 1), (' ', 1)] что после ма

counter.all_branches()                          # [(['м', 'а', 'м'], 1), (['м', 'а', ' '], 1),... все ветки

Вероятности в класса NGramsCounter вычисляются как простое отношение N1/N2.
Для более продвинутых способов следует использовать наследников от класса Markov.

Языковые модели содержат в себе counter, который наполняется в функциях add.
Но можно им передать и внешний counter, что удобно при использовании 
нескольких моделей:

lm = MarkovLaplas(beta=1, order=3, counter=counter)

print(lm.counter.prob("ру"))                    # 0    
print(lm.prob("ру"))                            # 0.125 = (0+1)/(1+7) 
print(lm.perplexity("мама умыла раму"))         # 4.886

Доступны следующие языковые модели:
    * Markov
    * MarkovLaplas
    * MarkovInterpolated
    * MarkovLaplasInterpolated
    * MarkovBackoff
"""
###############################################################################
#                                  class NGramsCounter
###############################################################################
class NGramsCounter:        
    """
    Подсчитывает число n-грамм с n=[1...order]. На их основе можно вычислять 
    условную вероятность данного слова перед которым шло (order-1) других слов. 
    Словами могут быть символы, строки или числа (индексы слов).
    """
    
    def __init__(self, order = 1):
        if order < 1:
            print("NGramsCounter: order must be > 0, got: ", order)
            order = 1
        
        self.order = order           # глубина дерева (макс. длина n-grams)        
        self.root  = {}              # корень дерева (наш словарь)
        self.ngrams= [0]*(order+1)   # число n-grams диной n=1,2,...,order
    #---------------------------------------------------------------------------                
    def add(self, lst):
        """
        Добавляем новые примеры из списка слов lst.
        Если слова это символы, то lst может быть строкой, иначе это список
        Структура дерева: node = {word: [count, node] }.
        Все возможные n-граммы (w_1...w_order) являются ветками дерева.
        Длина веток  меньше ли равна order. Все они начинаются от self.root.
        """            
        
        print("NGramsCounter::add", end='')
        for n in range(self.order):              # сколько каких n-grams добавилось
            self.ngrams[n+1] += max(len(lst) - n , 0)
            
        for i in range( len(lst) ):              # бежим по тексту           
            node = self.root
            for n, w in enumerate(lst[i: i+self.order]):       
                if w in node: node[w][0] += 1    # это слово уже было                  
                else:         node[w] = [1, {}]  # новое слово
                node = node[w][1]                # на следующий узел дерева
                
            if i % 64000 == 0:
                print('\r', " "*32,'\r', f"{i*100/len(lst):.1f}%", end='')
        print('\r', " "*32, '\r', end='')
    #---------------------------------------------------------------------------            
    def counts(self, lst):
        """
        Возвращает: N(w1...wn),   - число ngrams, где w_i - элементы списка lst
                    N(w1...wn-1), - число ngrams на 1 короче
                    узел          - ветки, выходящие из wn (если n=order, то это {})  
        Если len(lst) > order, то по её последним order символам        
        Если len(lst) == 1, то возвращает N(w), ngrams[1]                      
        """
        
        if len(lst) == 0:
            print("NGramsCounter::counts len(lst) must be > 0, got: ", len(lst))
            return 0, 0, self.root
        
        if len(lst) > self.order:
            lst = lst[-self.order : ]            # последние order символов
        
        node = self.root                         # корневой пучок веток                
        n_prv, n_cur  = self.ngrams[1], self.ngrams[1]
                        
        for i,w in enumerate(lst):               # по списку слов lst
            if w in node:                        # если слово есть в пучке
                n_prv = n_cur        
                n_cur = node[w][0]   
                node  = node[w][1]               # берём пучок веток из него
            else:
                if i==len(lst)-1:
                    return 0, n_cur, node        # последнее слово wn не нашли
                else:
                    return 0, 0,     node        # N(w1...wn-1) == 0 

        return n_cur, n_prv, node
    #---------------------------------------------------------------------------            
    def prob(self, lst, cond=True):
        """ 
        Вероятность P(lst[0],...,lst[n-2] => lst[n-1]), при n=1: P(lst[0]) 
        Если cond=False, то безусловная вероятность P(lst[0],...,lst[n-1])
        Если слова это символы, то lst может быть строкой, иначе это список.
        """        
            
        n_cur, n_prv, _ = self.counts(lst)
        if n_cur == 0:
            return 0                              # слово не нашли - вероятность 0
             
        return n_cur/(n_prv if cond else self.ngrams[len(lst)])        
        
    #---------------------------------------------------------------------------
    def branches(self, lst = []):
        """ 
        Cлова после списка слов lst в виде пар (слово, сколько): [(w1,cont1),...]
        Просто branches() - это словарь
        Если слова это символы, то lst может быть строкой, иначе это список.
        """
                
        if len(lst)==0:
            node = self.root
        else:
            n_cur, _, node = self.counts(lst)
            if n_cur == 0:
                return []
    
        res = [(w, node[w][0]) for w in node]      # пары (слово,вероятность)
        return sorted(res, key=lambda kv: -kv[1])  # сортированы по убыванию вероятности
        
    #---------------------------------------------------------------------------
    def unique(self, n = None):
        """
        Число уникальных n-грамм (число ветвей дерева длиной n)
        """
        if not n:
            return [ self.unique_rec(self.root, i) for i in range(1, self.order+1) ]
        return self.unique_rec(self.root, n)
    
    #---------------------------------------------------------------------------
    def unique_rec(self, node, n):
        if n <= 0:
            return 1
        s = 0
        for w in node:
            s += self.unique_rec(node[w][1], n-1)
        return s
                       
    #---------------------------------------------------------------------------                       
    def all_branches(self):
        """
        Возвращает все ветви дерева node с их числом появления в тексте.
        """        
        branches = []
        self.all_branches_rec(self.root,  [],  0, branches)
        return branches
    
    #---------------------------------------------------------------------------    
    def all_branches_rec(self,                
                         node,         # текущий узел дерева
                         cur,          # текущая к node ветка при спуске вниз
                         n_cur,        # число таких веток было в тексте
                         branches):    # все ветви дерева от корня                       
        if not node:
            branches.append( (cur, n_cur) )                        
            
        for n in node:                  # по всем детям из узла
            nxt = cur[:] + [n]          # делаем копию пути и добавляем ребёнка
            self.all_branches_rec(node[n][1], nxt, node[n][0], branches)
            
###############################################################################
#                                class LanguageModel
###############################################################################
            
class LanguageModel:
    """ 
    Абстрактный класс про-родитель языковых моделей. Умеет считать перплексию. 
    """
    def __init__(self, order,  counter = None):
        self.order   = order                      # число n-грамм
        self.counter = NGramsCounter(order) if not counter else counter        

    #---------------------------------------------------------------------------    
    def add(self, text):
        self.counter.add(text)    
    
    #---------------------------------------------------------------------------    
    def prob(self, lst):
        """ 
        Условная вероятность получить после слов lst[:-1] слово lst[-1]  
        """
        return self.counter.prob(lst)
        
    #---------------------------------------------------------------------------
    def perplexity(self,  text):
        """ 
        Вычисляет перплексию текста text для данной модели
        """
        
        order = self.order 
        s, num, eps = 0, len(text)-order, 1e-10
        for i in range(order, len(text)):                       
            p = self.prob(text[i-order:i])        # вероятность P(w1...w{n-1} => wn)
            if p < eps :
                print("LanguageModel::perplexity: prob:", p, i, text[i-order:i])
                return 0
            s += math.log(p)                      # произведение вероятностей            
            
        return math.exp(-s/num)                   # preplexity   
        
    #---------------------------------------------------------------------------        
    def generate(self, lst, num=1, insert=""):
        """ 
        Получить после списка слов lst следующее предсказываемое слово.
        Для слов (не букв) стоит положить insert = " " 
        """
        
        st = []
        for n in range(num):
            lst = lst[-self.order+1:]             # order - 1 слов
                
            vocab = self.counter.branches()       # все слова словаря
            probs = np.array([self.prob(list(lst) + [w]) for w,_ in vocab])             
            if abs(sum(probs) - 1) > 1e-8:
                if 1-sum(probs)  < 1e-4 and sum(probs) <= 1.0:
                    probs /= probs.sum()
                else:
                    print("Wrong context P(contex)=0 was created. sum(prob)!=1", sum(probs))
                    return st
            
            i = np.random.choice(len(probs), 1, p=probs)[0]            
            lst += vocab[i][0]
            st  += vocab[i][0] + insert
        return "".join(st)
 
###############################################################################
#                              class Markov
###############################################################################


class Markov(LanguageModel):
    """ 
    Простая марковская модель
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)                   

    def prob(self, lst):
        """ 
        Вероятность получить после списка слов lst[:-1] слово lst[-1]  
        """        
        return self.counter.prob(lst)

###############################################################################
#                              class MarkovLaplas
###############################################################################


class MarkovLaplas(LanguageModel):
    """ 
    Сглаживание Лапласа
    """

    def __init__(self, beta, *args, **kwargs):
        super().__init__(*args, **kwargs)                   
        self.beta = beta

    def prob(self, lst):
        """ 
        Вероятность получить после списка слов lst[:-1] слово lst[-1]  
        """        
        
        n_cur, n_prv, n = self.counter.counts(lst)                     
        return (n_cur + self.beta)/(n_prv + self.beta*len(self.counter.root) )        


###############################################################################
#                              class MarkovInterpolated 
###############################################################################


class MarkovInterpolated(LanguageModel):
    """ 
    Наследник LanguageModel, реализующий марковскую модель с интерполяцией 
    """
        
    def __init__(self, beta, minN, *args, **kwargs):
        super().__init__(*args, **kwargs)                   
        self.beta = beta                       # коэффициент интерполяции
        if minN < 1: minN = 1
        self.minN = minN                       # сколько должно быть примеров  
                        
    def prob(self, lst):
        """ 
        Вероятность получить после списка слов lst[:-1] слово lst[-1]  
        """        
        
        p, norm, coef = 0, 0, 1;
        for i in range(self.order):     
            n_cur, n_prv, _ = self.counter.counts(lst[-(self.order-i): ])
            if n_prv >= self.minN:                
                p    += coef * (n_cur / n_prv)
                norm += coef       
            coef *= self.beta                      
        return p/norm                        

###############################################################################
#                              class MarkovLaplasInterpolated 
###############################################################################


class MarkovLaplasInterpolated(LanguageModel):
    """ 
    Наследник LanguageModel, реализующий марковскую модель с интерполяцией 
    """
        
    def __init__(self, beta, gamma, *args, **kwargs):
        super().__init__(*args, **kwargs)                   
        self.beta    = beta                       # коэффициент интерполяции
        self.gamma   = gamma
                        
    def prob(self, lst):
        """ 
        Вероятность получить после списка слов lst[:-1] слово lst[-1]  
        """        
        
        p, norm, coef, voc = 0, 0, 1, len(self.counter.root);
        for i in range(self.order):     
            n_cur, n_prv, _ = self.counter.counts(lst[-(self.order-i): ])            
            p    += coef * (n_cur + self.gamma) / (n_prv + self.gamma*voc)
            norm += coef       
            coef *= self.beta                      
        return p/norm                        
    
###############################################################################
#                                 class MarkovBackoff
###############################################################################

class MarkovBackoff(LanguageModel):
    """ 
    Наследник LanguageModel, реализующий марковскую модель с возвратом.
    Внимание! Сумма вероятностей может быть больше единицы.
    """
        
    def __init__(self, beta, *args, **kwargs):
        super().__init__(*args, **kwargs)                   
                        
    def prob(self, lst):
        """ 
        Вероятность получить после списка слов lst[:-1] слово lst[-1]  
        """        
        
        for i in range(self.order):                    
            p = self.counter.prob(lst[-(self.order-i): ])             
            if p > 0:
                return p
        print("Markov.prob: OOV")
        return 0                             # ? слова нет!
